;; Generated package description from speedrect.el  -*- no-byte-compile: t -*-
(define-package "speedrect" "0.6.0.20241209.32904" "Fast modal rectangle commands" '((emacs "29.1") (compat "30")) :commit "0a97e82913a3834058b2365402ad8e96a06d2259" :authors '(("JD Smith" . "jdtsmith+elpa@gmail.com")) :maintainer '("JD Smith" . "jdtsmith+elpa@gmail.com") :keywords '("convenience") :url "https://github.com/jdtsmith/speedrect")
